package org.apache.bcel.util;

import java.util.Arrays;
import java.util.function.Function;
import java.nio.file.OpenOption;
import java.io.FileInputStream;
import java.net.MalformedURLException;
import java.util.zip.ZipEntry;
import java.util.Objects;
import java.util.Locale;
import java.util.Vector;
import java.util.Enumeration;
import java.net.URL;
import java.io.InputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.util.zip.ZipFile;
import java.util.StringTokenizer;
import java.util.Iterator;
import java.nio.file.Path;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.io.File;
import java.util.List;
import java.io.FilenameFilter;
import java.io.Closeable;

public class ClassPath implements Closeable
{
    private static final FilenameFilter ARCHIVE_FILTER;
    private static final FilenameFilter MODULES_FILTER;
    public static final ClassPath SYSTEM_CLASS_PATH;
    private final String classPath;
    private ClassPath parent;
    private final AbstractPathEntry[] paths;
    
    private static void addJdkModules(final String javaHome, final List<String> list) {
        String modulesPath = System.getProperty("java.modules.path");
        if (modulesPath == null || modulesPath.trim().isEmpty()) {
            modulesPath = javaHome + File.separator + "jmods";
        }
        final File modulesDir = new File(modulesPath);
        if (modulesDir.exists()) {
            final String[] modules = modulesDir.list(ClassPath.MODULES_FILTER);
            if (modules != null) {
                for (final String module : modules) {
                    list.add(modulesDir.getPath() + File.separatorChar + module);
                }
            }
        }
    }
    
    public static String getClassPath() {
        final String classPathProp = System.getProperty("java.class.path");
        final String bootClassPathProp = System.getProperty("sun.boot.class.path");
        final String extDirs = System.getProperty("java.ext.dirs");
        final String javaHome = System.getProperty("java.home");
        final List<String> list = new ArrayList<String>();
        final Path modulesPath = Paths.get(javaHome, new String[0]).resolve("lib/modules");
        if (Files.exists(modulesPath, new LinkOption[0]) && Files.isRegularFile(modulesPath, new LinkOption[0])) {
            list.add(modulesPath.toAbsolutePath().toString());
        }
        addJdkModules(javaHome, list);
        getPathComponents(classPathProp, list);
        getPathComponents(bootClassPathProp, list);
        final List<String> dirs = new ArrayList<String>();
        getPathComponents(extDirs, dirs);
        for (final String d : dirs) {
            final File extDir = new File(d);
            final String[] extensions = extDir.list(ClassPath.ARCHIVE_FILTER);
            if (extensions != null) {
                for (final String extension : extensions) {
                    list.add(extDir.getPath() + File.separatorChar + extension);
                }
            }
        }
        final StringBuilder buf = new StringBuilder();
        String separator = "";
        for (final String path : list) {
            buf.append(separator);
            separator = File.pathSeparator;
            buf.append(path);
        }
        return buf.toString().intern();
    }
    
    private static void getPathComponents(final String path, final List<String> list) {
        if (path != null) {
            final StringTokenizer tokenizer = new StringTokenizer(path, File.pathSeparator);
            while (tokenizer.hasMoreTokens()) {
                final String name = tokenizer.nextToken();
                final File file = new File(name);
                if (file.exists()) {
                    list.add(name);
                }
            }
        }
    }
    
    static String packageToFolder(final String name) {
        return name.replace('.', '/');
    }
    
    @Deprecated
    public ClassPath() {
        this(getClassPath());
    }
    
    public ClassPath(final ClassPath parent, final String classPath) {
        this(classPath);
        this.parent = parent;
    }
    
    public ClassPath(final String classPath) {
        this.classPath = classPath;
        final List<AbstractPathEntry> list = new ArrayList<AbstractPathEntry>();
        final StringTokenizer tokenizer = new StringTokenizer(classPath, File.pathSeparator);
        while (tokenizer.hasMoreTokens()) {
            final String path = tokenizer.nextToken();
            if (!path.isEmpty()) {
                final File file = new File(path);
                try {
                    if (!file.exists()) {
                        continue;
                    }
                    if (file.isDirectory()) {
                        list.add(new Dir(path));
                    }
                    else if (path.endsWith(".jmod")) {
                        list.add(new Module(new ZipFile(file)));
                    }
                    else if (path.endsWith(ModularRuntimeImage.MODULES_PATH)) {
                        list.add(new JrtModules(ModularRuntimeImage.MODULES_PATH));
                    }
                    else {
                        list.add(new Jar(new ZipFile(file)));
                    }
                }
                catch (final IOException e) {
                    if (!path.endsWith(".zip") && !path.endsWith(".jar")) {
                        continue;
                    }
                    System.err.println("CLASSPATH component " + file + ": " + e);
                }
            }
        }
        list.toArray(this.paths = new AbstractPathEntry[list.size()]);
    }
    
    @Override
    public void close() throws IOException {
        if (this.paths != null) {
            for (final AbstractPathEntry path : this.paths) {
                path.close();
            }
        }
    }
    
    @Override
    public boolean equals(final Object o) {
        if (o instanceof ClassPath) {
            final ClassPath cp = (ClassPath)o;
            return this.classPath.equals(cp.toString());
        }
        return false;
    }
    
    public byte[] getBytes(final String name) throws IOException {
        return this.getBytes(name, ".class");
    }
    
    public byte[] getBytes(final String name, final String suffix) throws IOException {
        DataInputStream dis = null;
        try (InputStream inputStream = this.getInputStream(name, suffix)) {
            if (inputStream == null) {
                throw new IOException("Couldn't find: " + name + suffix);
            }
            dis = new DataInputStream(inputStream);
            final byte[] bytes = new byte[inputStream.available()];
            dis.readFully(bytes);
            return bytes;
        }
        finally {
            if (dis != null) {
                dis.close();
            }
        }
    }
    
    public ClassFile getClassFile(final String name) throws IOException {
        return this.getClassFile(name, ".class");
    }
    
    public ClassFile getClassFile(final String name, final String suffix) throws IOException {
        ClassFile cf = null;
        if (this.parent != null) {
            cf = this.parent.getClassFileInternal(name, suffix);
        }
        if (cf == null) {
            cf = this.getClassFileInternal(name, suffix);
        }
        if (cf != null) {
            return cf;
        }
        throw new IOException("Couldn't find: " + name + suffix);
    }
    
    private ClassFile getClassFileInternal(final String name, final String suffix) {
        for (final AbstractPathEntry path : this.paths) {
            final ClassFile cf = path.getClassFile(name, suffix);
            if (cf != null) {
                return cf;
            }
        }
        return null;
    }
    
    public InputStream getInputStream(final String name) throws IOException {
        return this.getInputStream(packageToFolder(name), ".class");
    }
    
    public InputStream getInputStream(final String name, final String suffix) throws IOException {
        InputStream inputStream = null;
        try {
            inputStream = this.getClass().getClassLoader().getResourceAsStream(name + suffix);
        }
        catch (final Exception ex) {}
        if (inputStream != null) {
            return inputStream;
        }
        return this.getClassFile(name, suffix).getInputStream();
    }
    
    public String getPath(String name) throws IOException {
        final int index = name.lastIndexOf(46);
        String suffix = "";
        if (index > 0) {
            suffix = name.substring(index);
            name = name.substring(0, index);
        }
        return this.getPath(name, suffix);
    }
    
    public String getPath(final String name, final String suffix) throws IOException {
        return this.getClassFile(name, suffix).getPath();
    }
    
    public URL getResource(final String name) {
        for (final AbstractPathEntry path : this.paths) {
            final URL url;
            if ((url = path.getResource(name)) != null) {
                return url;
            }
        }
        return null;
    }
    
    public InputStream getResourceAsStream(final String name) {
        for (final AbstractPathEntry path : this.paths) {
            final InputStream is;
            if ((is = path.getResourceAsStream(name)) != null) {
                return is;
            }
        }
        return null;
    }
    
    public Enumeration<URL> getResources(final String name) {
        final Vector<URL> results = new Vector<URL>();
        for (final AbstractPathEntry path : this.paths) {
            final URL url;
            if ((url = path.getResource(name)) != null) {
                results.add(url);
            }
        }
        return results.elements();
    }
    
    @Override
    public int hashCode() {
        if (this.parent != null) {
            return this.classPath.hashCode() + this.parent.hashCode();
        }
        return this.classPath.hashCode();
    }
    
    @Override
    public String toString() {
        if (this.parent != null) {
            return this.parent + File.pathSeparator + this.classPath;
        }
        return this.classPath;
    }
    
    static {
        ARCHIVE_FILTER = ((dir, name) -> {
            name = name.toLowerCase(Locale.ENGLISH);
            return name.endsWith(".zip") || name.endsWith(".jar");
        });
        MODULES_FILTER = ((dir, name) -> {
            name = name.toLowerCase(Locale.ENGLISH);
            return name.endsWith(".jmod");
        });
        SYSTEM_CLASS_PATH = new ClassPath(getClassPath());
    }
    
    private abstract static class AbstractPathEntry implements Closeable
    {
        abstract ClassFile getClassFile(final String p0, final String p1);
        
        abstract URL getResource(final String p0);
        
        abstract InputStream getResourceAsStream(final String p0);
    }
    
    private abstract static class AbstractZip extends AbstractPathEntry
    {
        private final ZipFile zipFile;
        
        AbstractZip(final ZipFile zipFile) {
            this.zipFile = Objects.requireNonNull(zipFile, "zipFile");
        }
        
        @Override
        public void close() throws IOException {
            if (this.zipFile != null) {
                this.zipFile.close();
            }
        }
        
        @Override
        ClassFile getClassFile(final String name, final String suffix) {
            final ZipEntry entry = this.zipFile.getEntry(this.toEntryName(name, suffix));
            if (entry == null) {
                return null;
            }
            return new ClassFile() {
                @Override
                public String getBase() {
                    return AbstractZip.this.zipFile.getName();
                }
                
                @Override
                public InputStream getInputStream() throws IOException {
                    return AbstractZip.this.zipFile.getInputStream(entry);
                }
                
                @Override
                public String getPath() {
                    return entry.toString();
                }
                
                @Override
                public long getSize() {
                    return entry.getSize();
                }
                
                @Override
                public long getTime() {
                    return entry.getTime();
                }
            };
        }
        
        @Override
        URL getResource(final String name) {
            final ZipEntry entry = this.zipFile.getEntry(name);
            try {
                URL url;
                if (entry != null) {
                    final StringBuilder sb;
                    url = new URL(sb.append("jar:file:").append(this.zipFile.getName()).append("!/").append(name).toString());
                    sb = new StringBuilder();
                }
                else {
                    url = null;
                }
                return url;
            }
            catch (final MalformedURLException e) {
                return null;
            }
        }
        
        @Override
        InputStream getResourceAsStream(final String name) {
            final ZipEntry entry = this.zipFile.getEntry(name);
            try {
                return (entry != null) ? this.zipFile.getInputStream(entry) : null;
            }
            catch (final IOException e) {
                return null;
            }
        }
        
        protected abstract String toEntryName(final String p0, final String p1);
        
        @Override
        public String toString() {
            return this.zipFile.getName();
        }
    }
    
    private static class Dir extends AbstractPathEntry
    {
        private final String dir;
        
        Dir(final String d) {
            this.dir = d;
        }
        
        @Override
        public void close() throws IOException {
        }
        
        @Override
        ClassFile getClassFile(final String name, final String suffix) {
            final File file = new File(this.dir + File.separatorChar + name.replace('.', File.separatorChar) + suffix);
            return file.exists() ? new ClassFile() {
                @Override
                public String getBase() {
                    return Dir.this.dir;
                }
                
                @Override
                public InputStream getInputStream() throws IOException {
                    return new FileInputStream(file);
                }
                
                @Override
                public String getPath() {
                    try {
                        return file.getCanonicalPath();
                    }
                    catch (final IOException e) {
                        return null;
                    }
                }
                
                @Override
                public long getSize() {
                    return file.length();
                }
                
                @Override
                public long getTime() {
                    return file.lastModified();
                }
            } : null;
        }
        
        @Override
        URL getResource(final String name) {
            final File file = this.toFile(name);
            try {
                return file.exists() ? file.toURI().toURL() : null;
            }
            catch (final MalformedURLException e) {
                return null;
            }
        }
        
        @Override
        InputStream getResourceAsStream(final String name) {
            final File file = this.toFile(name);
            try {
                return file.exists() ? new FileInputStream(file) : null;
            }
            catch (final IOException e) {
                return null;
            }
        }
        
        private File toFile(final String name) {
            return new File(this.dir + File.separatorChar + name.replace('/', File.separatorChar));
        }
        
        @Override
        public String toString() {
            return this.dir;
        }
    }
    
    private static class Jar extends AbstractZip
    {
        Jar(final ZipFile zip) {
            super(zip);
        }
        
        @Override
        protected String toEntryName(final String name, final String suffix) {
            return ClassPath.packageToFolder(name) + suffix;
        }
    }
    
    private static class JrtModule extends AbstractPathEntry
    {
        private final Path modulePath;
        
        public JrtModule(final Path modulePath) {
            this.modulePath = Objects.requireNonNull(modulePath, "modulePath");
        }
        
        @Override
        public void close() throws IOException {
        }
        
        @Override
        ClassFile getClassFile(final String name, final String suffix) {
            final Path resolved = this.modulePath.resolve(ClassPath.packageToFolder(name) + suffix);
            if (Files.exists(resolved, new LinkOption[0])) {
                return new ClassFile() {
                    @Override
                    public String getBase() {
                        return Objects.toString(resolved.getFileName(), null);
                    }
                    
                    @Override
                    public InputStream getInputStream() throws IOException {
                        return Files.newInputStream(resolved, new OpenOption[0]);
                    }
                    
                    @Override
                    public String getPath() {
                        return resolved.toString();
                    }
                    
                    @Override
                    public long getSize() {
                        try {
                            return Files.size(resolved);
                        }
                        catch (final IOException e) {
                            return 0L;
                        }
                    }
                    
                    @Override
                    public long getTime() {
                        try {
                            return Files.getLastModifiedTime(resolved, new LinkOption[0]).toMillis();
                        }
                        catch (final IOException e) {
                            return 0L;
                        }
                    }
                };
            }
            return null;
        }
        
        @Override
        URL getResource(final String name) {
            final Path resovled = this.modulePath.resolve(name);
            try {
                URL url;
                if (Files.exists(resovled, new LinkOption[0])) {
                    final StringBuilder sb;
                    url = new URL(sb.append("jrt:").append(this.modulePath).append("/").append(name).toString());
                    sb = new StringBuilder();
                }
                else {
                    url = null;
                }
                return url;
            }
            catch (final MalformedURLException e) {
                return null;
            }
        }
        
        @Override
        InputStream getResourceAsStream(final String name) {
            try {
                return Files.newInputStream(this.modulePath.resolve(name), new OpenOption[0]);
            }
            catch (final IOException e) {
                return null;
            }
        }
        
        @Override
        public String toString() {
            return this.modulePath.toString();
        }
    }
    
    private static class JrtModules extends AbstractPathEntry
    {
        private final ModularRuntimeImage modularRuntimeImage;
        private final JrtModule[] modules;
        
        public JrtModules(final String path) throws IOException {
            this.modularRuntimeImage = new ModularRuntimeImage();
            this.modules = this.modularRuntimeImage.list(path).stream().map((Function<? super Object, ?>)JrtModule::new).toArray(JrtModule[]::new);
        }
        
        @Override
        public void close() throws IOException {
            if (this.modules != null) {
                for (final JrtModule module : this.modules) {
                    module.close();
                }
            }
            if (this.modularRuntimeImage != null) {
                this.modularRuntimeImage.close();
            }
        }
        
        @Override
        ClassFile getClassFile(final String name, final String suffix) {
            for (final JrtModule module : this.modules) {
                final ClassFile classFile = module.getClassFile(name, suffix);
                if (classFile != null) {
                    return classFile;
                }
            }
            return null;
        }
        
        @Override
        URL getResource(final String name) {
            for (final JrtModule module : this.modules) {
                final URL url = module.getResource(name);
                if (url != null) {
                    return url;
                }
            }
            return null;
        }
        
        @Override
        InputStream getResourceAsStream(final String name) {
            for (final JrtModule module : this.modules) {
                final InputStream inputStream = module.getResourceAsStream(name);
                if (inputStream != null) {
                    return inputStream;
                }
            }
            return null;
        }
        
        @Override
        public String toString() {
            return Arrays.toString(this.modules);
        }
    }
    
    private static class Module extends AbstractZip
    {
        Module(final ZipFile zip) {
            super(zip);
        }
        
        @Override
        protected String toEntryName(final String name, final String suffix) {
            return "classes/" + ClassPath.packageToFolder(name) + suffix;
        }
    }
    
    public interface ClassFile
    {
        String getBase();
        
        InputStream getInputStream() throws IOException;
        
        String getPath();
        
        long getSize();
        
        long getTime();
    }
}
